//=============================================================================
//
//		Copyright (C) 1995 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//
//=============================================================================
//-----------------------------------------------------------------------------
// DirWindow.m
//
//		Manages window which displays directory listing.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id$
// $Log$
//-----------------------------------------------------------------------------

#import "DirWindow.h"
#import <misckit/MiscTableScroll.h>
#import <misckit/MiscTableCell.h>
#import <errno.h>
#import <grp.h>
#import <pwd.h>
#import <appkit/NXImage.h>
#import <appkit/workspaceRequest.h>
#import <sys/dir.h>

#ifdef _POSIX_SOURCE
# define GET_CURR_DIR(_BF_,_SZ_)		getcwd(_BF_,_SZ_)
#else
# define GET_CURR_DIR(_BF_,_SZ_)		getwd(_BF_)
#endif

enum
	{
	ICON_SLOT,
	NAME_SLOT,
	SIZE_SLOT,
	MODIFIED_SLOT,
	PERMS_SLOT,
	OWNER_SLOT,
	GROUP_SLOT,
	HARDLINKS_SLOT,
	SOFTLINK_SLOT,
	MESSAGE_SLOT
	};
#define MAX_SLOT (MESSAGE_SLOT + 1)

#define CASCADE_MAX 10
static int cascadeCounter = 0;
static NXCoord cascadeOriginX;
static NXCoord cascadeOriginY;
static NXCoord cascadeDeltaX;
static NXCoord cascadeDeltaY;

static NXColor defaultColor;
static List* openDirs = 0;


//-----------------------------------------------------------------------------
// normalize_path
//-----------------------------------------------------------------------------
static void normalize_path( char *buff, size_t sz )
	{
	char save_dir[ FILENAME_MAX + 1 ];
	char tmp[ FILENAME_MAX + 1 ];

	if (buff != 0 && sz > 0)
		{
		if (GET_CURR_DIR( save_dir, sizeof(save_dir) ) != 0)
			{
			if (chdir( buff ) == 0)
				{
				if (GET_CURR_DIR( tmp, sizeof(tmp) ) != 0)
					{
					int const tmp_len = strlen( tmp );
					if (tmp_len < sz)
						strcpy( buff, tmp );
					else
						{
						strncpy( buff, tmp, sz );
						buff[ sz - 1 ] = '\0';
						}
					}
				chdir( save_dir );
				}
			}
		}
	}


//-----------------------------------------------------------------------------
// fsize
//-----------------------------------------------------------------------------
static void fsize( size_t n, id cell )
	{
	[[cell setAlignment: NX_RIGHTALIGNED] setIntValue: n];
	}


//-----------------------------------------------------------------------------
// mdate
//-----------------------------------------------------------------------------
static void mdate( time_t t, id cell )
	{
	int const BUFF_SIZE = 64;
	char buff[ BUFF_SIZE ];
	strftime( buff, BUFF_SIZE, "%m/%d/%y %H:%M", localtime(&t) );
	[cell setStringValue:buff];
	[cell setTag: (int) t ];
	}


//-----------------------------------------------------------------------------
// perms
//-----------------------------------------------------------------------------
static void perms( unsigned int mode, id cell )
	{
	char buff[ 16 ];
	char* s;
	int ftype;
	int c;

	s = buff;

	ftype = (mode & S_IFMT);
	if (ftype == S_IFDIR)		c = 'd';
	else if (ftype == S_IFCHR)	c = 'c';
	else if (ftype == S_IFBLK)	c = 'b';
	else if (ftype == S_IFREG)	c = '-';
	else if (ftype == S_IFLNK)	c = 'l';
	else if (ftype == S_IFSOCK) c = 's';
	else c = '-';
	*s++ = c;
	*s++ = (mode & 0400) ? 'r' : '-';
	*s++ = (mode & 0200) ? 'w' : '-';
	*s++ = (mode & S_ISUID) ? 's' : (mode & 0100) ? 'x' : '-';
	*s++ = (mode & 0040) ? 'r' : '-';
	*s++ = (mode & 0020) ? 'w' : '-';
	*s++ = (mode & S_ISGID) ? 's' : (mode & 0010) ? 'x' : '-';
	*s++ = (mode & 0004) ? 'r' : '-';
	*s++ = (mode & 0002) ? 'w' : '-';
	*s++ = (mode & S_ISVTX) ? 't' : (mode & 0001) ? 'x' : '-';
	*s = '\0';

	[cell setStringValue: buff];
	[cell setTag: mode];
	}


//-----------------------------------------------------------------------------
// owner
//-----------------------------------------------------------------------------
static void owner( unsigned int uid, id cell )
	{
	struct passwd* pw = getpwuid( uid );
	if (pw)
		[cell setStringValue: pw->pw_name];
	else
		[cell setIntValue: uid];
	}


//-----------------------------------------------------------------------------
// group
//-----------------------------------------------------------------------------
static void group( unsigned int gid, id cell )
	{
	struct group* gr = getgrgid( gid );
	if (gr)
		[cell setStringValue: gr->gr_name];
	else
		[cell setIntValue: gid];
	}


//-----------------------------------------------------------------------------
// image
//-----------------------------------------------------------------------------
static void image( char const* name, id cell, NXCoord w, NXCoord h )
	{
	NXCoord s = (w < h ? w : h) - 1.0;
	NXSize sz = { s, s };
	NXImage* i = [[Application workspace] getIconForFile: name];
	[i setScalable: YES];
	[i setSize: &sz];
	[cell setImage: i];
	}



//=============================================================================
// IMPLEMENTATION
//=============================================================================
@implementation DirWindow

//-----------------------------------------------------------------------------
// + initCascader
//-----------------------------------------------------------------------------
+ (void) initCascader
	{
	NXSize screenSize;
	[NXApp getScreenSize: &screenSize];
	cascadeOriginX = screenSize.width / 4;
	cascadeOriginY = screenSize.height - 20;
	cascadeDeltaX = 20;
	cascadeDeltaY = 20;
	}


//-----------------------------------------------------------------------------
// + initialize
//-----------------------------------------------------------------------------
+ initialize
	{
	if (self == [DirWindow class])
		{
		[self initCascader];
		defaultColor = NX_COLORLTGRAY;
		openDirs = [[List alloc] init];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// - cascade
//-----------------------------------------------------------------------------
- (void) cascade
	{
	NXCoord top,left;
	left = cascadeOriginX + (cascadeDeltaX * cascadeCounter);
	top	 = cascadeOriginY - (cascadeDeltaY * cascadeCounter);
	[window moveTopLeftTo: left : top ];
	if (++cascadeCounter >= CASCADE_MAX)
		cascadeCounter = 0;
	}


//-----------------------------------------------------------------------------
// - updateButtons
//-----------------------------------------------------------------------------
- (void) updateButtons
	{
	BOOL enable = NO;
	if ([scroll numSelectedRows] == 1)
		{
		MiscCoord_P const row = [scroll selectedRow];
		id cell = [scroll cellAt:row:PERMS_SLOT];
		unsigned int const mode = (unsigned int) [cell tag];
		enable = ((mode & S_IFMT) == S_IFDIR);
		}
	if (enable != [cdButton isEnabled])
		[cdButton setEnabled: enable];
	}


//-----------------------------------------------------------------------------
// - highlightRow:
//-----------------------------------------------------------------------------
- (void) highlightRow: (int)r
	{
	int i;
	for (i = MAX_SLOT; i-- >= 0; )
		{
		id c = [scroll cellAt:r:i];
		if ([c respondsTo: @selector(setBackgroundColor:)])
			[c setBackgroundColor: NX_COLORCYAN];
		}
	}


//-----------------------------------------------------------------------------
// - fillScroll
//-----------------------------------------------------------------------------
- (void) fillScroll
	{
	size_t total_bytes = 0;
	int dirlen;
	char namebuff[ FILENAME_MAX + 1 ];
	DIR* dirp;

	dirlen = strlen( path );
	memcpy( namebuff, path, dirlen );

	if (dirlen == 0 || namebuff[ dirlen - 1 ] != '/')
		namebuff[ dirlen++ ] = '/';

	[window disableDisplay];

	[scroll empty];
	if ((dirp = opendir( path )) == 0)
		{
		NXRunAlertPanel( "Can't Open",
				"Cannot open directory, %s\n%d:%s", "OK",0,0,
				path, errno, strerror(errno) );
		}
	else
		{
		int rc;
		int row = -1;
		struct stat st;
		struct direct const* dp;

		while ((dp = readdir(dirp)) != 0)
			{			// Do not include the "." (self) entry.
			BOOL const dot_name = (dp->d_name[0] == '.');
			if ((dp->d_namlen > 1 || !dot_name) &&		// exclude "."
				(showHidden || !dot_name ||
					(dp->d_namlen == 2 && dp->d_name[1] == '.')))
				{
				[scroll addRow]; row++;
				[[scroll cellAt:row:NAME_SLOT] setStringValue:dp->d_name];
				strcpy( namebuff + dirlen, dp->d_name );
				if ((rc = lstat( namebuff, &st )) == 0)
					{
					if ((st.st_mode & S_IFMT) == S_IFLNK)		// soft-link?
						{
						char symbuf[ FILENAME_MAX + 1 ];
						int cc;
						cc = readlink( namebuff, symbuf, FILENAME_MAX );
						if (cc >= 0)
							{
							symbuf[ cc ] = '\0';
							[[scroll cellAt:row:SOFTLINK_SLOT]
												setStringValue:symbuf];
							rc = stat( namebuff, &st );
							}
						else
							rc = -1;
						}
					}
				if (rc == 0)
					{
					total_bytes += st.st_size;
					image( namebuff, [scroll cellAt:row:ICON_SLOT],
								[scroll colSize:ICON_SLOT],
								[scroll rowSize:ICON_SLOT] );
					fsize( st.st_size, [scroll cellAt:row:SIZE_SLOT] );
					mdate( st.st_mtime,[scroll cellAt:row:MODIFIED_SLOT] );
					perms( st.st_mode, [scroll cellAt:row:PERMS_SLOT] );
					owner( st.st_uid, [scroll cellAt:row:OWNER_SLOT] );
					group( st.st_gid, [scroll cellAt:row:GROUP_SLOT] );
					[[scroll cellAt:row:HARDLINKS_SLOT]
								setIntValue:st.st_nlink];
					[[scroll cellAt:row:MESSAGE_SLOT]
								setStringValueNoCopy:""];
					if (highlightDirs && (st.st_mode & S_IFMT) == S_IFDIR)
						[self highlightRow: row];
					}
				else // (rc != 0)
					{
					int slot;
					[[scroll cellAt:row:MESSAGE_SLOT]
								setStringValue: strerror(errno)];
					for (slot = SIZE_SLOT; slot < SOFTLINK_SLOT; slot++)
						[[scroll cellAt:row:slot] setStringValueNoCopy:"?"];
					}
				}
			}
		}
	if ([scroll autoSortRows])
		[scroll sortRows];
	[scroll sizeToCells];
	sprintf( namebuff, "%d files   %lu bytes", [scroll numRows], total_bytes );
	[countField setStringValue:namebuff];
	[self updateButtons];
	[[window reenableDisplay] display];
	}


//-----------------------------------------------------------------------------
// - setPath:
//-----------------------------------------------------------------------------
- (void) setPath:(char const*)dirname
	{
	[window setTitle:dirname];
	if (path != 0) free(path);
	if (dirname == 0) dirname = NXHomeDirectory();
	if (dirname == 0) dirname = "/";
	strcpy( (path = (char*)malloc( strlen(dirname) + 1 )), dirname );
	}


//-----------------------------------------------------------------------------
// - load:
//-----------------------------------------------------------------------------
- (void) load:(char const*)dirname
	{
	[self setPath:dirname];
	[self fillScroll];
	}


//-----------------------------------------------------------------------------
// - open:
//-----------------------------------------------------------------------------
- open:sender
	{
	if ([scroll hasRowSelection])
		{
		int path_len;
		char buff[ FILENAME_MAX + 1 ];
		int i;
		MiscIntList* list = [[MiscIntList alloc] init];

		strcpy( buff, path );
		path_len = strlen( buff );
		if (path_len == 0 || buff[path_len - 1] != '/')
			{
			buff[ path_len++ ] = '/';
			buff[ path_len ] = '\0';
			}
			
		[scroll selectedRows: list];
		for (i = [list count]; i-- > 0; )
			{
			unsigned int mode;
			id cell;
			int row = [list intAt:i];

			strcpy( buff + path_len,
						[[scroll cellAt: row : NAME_SLOT] stringValue] );

			cell = [scroll cellAt: row : PERMS_SLOT];
			mode = (unsigned int)[cell tag];
			if ((mode & S_IFMT) == S_IFDIR)
				[[self class] launchDir: buff];
			else
				[[Application workspace] openFile: buff];
			}
		[list free];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// - destroy:
//-----------------------------------------------------------------------------
- destroy:sender
	{
	if ([scroll hasRowSelection])
		{
		size_t total_len = 0;
		char* file_list;
		int i;
		MiscIntList* list = [[MiscIntList alloc] init];

		[scroll selectedRows: list];

		for (i = [list count]; i-- > 0; )
			{
			int row = [list intAt:i];
			char const* s = [[scroll cellAt:row:NAME_SLOT] stringValue];
			if (s != 0)
				total_len += strlen( s ) + 1;
			}

		file_list = (char*) malloc( total_len + 1 );
		
		total_len = 0;
		for (i = [list count]; i-- > 0; )
			{
			int row = [list intAt:i];
			char const* s = [[scroll cellAt:row:NAME_SLOT] stringValue];
			if (s != 0)
				{
				int len = strlen( s );
				strcpy( file_list + total_len, s );
				total_len += len;
				file_list[ total_len++ ] = '\t';
				}
			}
		file_list[ total_len ] = '\0';

		[[Application workspace]
				performFileOperation: WSM_DESTROY_OPERATION
						source: path
						destination: ""
						files: file_list
						options: ""];

		free( file_list );
		[list free];
		[self fillScroll];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// - updatePressed:
//-----------------------------------------------------------------------------
- updatePressed:sender
	{
	[self fillScroll];
	return self;
	}


//-----------------------------------------------------------------------------
// - cdPressed:
//-----------------------------------------------------------------------------
- cdPressed:sender
	{
	if ([scroll numSelectedRows] == 1)
		{
		MiscCoord_P const row = [scroll selectedRow];
		id cell = [scroll cellAt:row:PERMS_SLOT];
		unsigned int const mode = (unsigned int) [cell tag];
		if ((mode & S_IFMT) == S_IFDIR)
			{
			char buff[ FILENAME_MAX * 2 + 1 ];
			cell = [scroll cellAt:row:NAME_SLOT];
			strcat( strcat( strcpy( buff, path ), "/" ), [cell stringValue] );
			normalize_path( buff, sizeof(buff) );
			[self load:buff];
			}
		}
	return self;
	}


//-----------------------------------------------------------------------------
// - autoSortClick:
//-----------------------------------------------------------------------------
- autoSortClick:sender
	{
	BOOL const switchState = [autoSortSwitch state];
	if ([scroll autoSortRows] != switchState)
		{
		[scroll setAutoSortRows:switchState];
		if (switchState)
			[scroll sortRows];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// - hiddenFilesClick:
//-----------------------------------------------------------------------------
- hiddenFilesClick:sender
	{
	if (showHidden != [hiddenFilesSwitch state])
		{
		showHidden = [hiddenFilesSwitch state];
		[self fillScroll];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// - highlightClick:
//-----------------------------------------------------------------------------
- highlightClick:sender
	{
	if (highlightDirs != [highlightSwitch state])
		{
		highlightDirs = [highlightSwitch state];
		[self fillScroll];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// - didClick:
//-----------------------------------------------------------------------------
- didClick:sender
	{
	[self updateButtons];
	return self;
	}


//-----------------------------------------------------------------------------
// - didDoubleClick:
//-----------------------------------------------------------------------------
- didDoubleClick:sender
	{
	[self open:sender];
	return self;
	}


//-----------------------------------------------------------------------------
// - free
//-----------------------------------------------------------------------------
- free
	{
	[window close];
	[window free];
	if (path != 0)
		free( path );
	return [super free];
	}


//-----------------------------------------------------------------------------
// - makeKeyAndOrderFront:
//-----------------------------------------------------------------------------
- makeKeyAndOrderFront:sender
	{
	[window makeKeyAndOrderFront:sender];
	return self;
	}


//-----------------------------------------------------------------------------
// - windowWillClose:
//-----------------------------------------------------------------------------
- windowWillClose:sender
	{
	[openDirs removeObject:self];
	[NXApp delayedFree: self];
	return self;
	}


//-----------------------------------------------------------------------------
// - setDaColors:
//-----------------------------------------------------------------------------
- (void) setDaColors: (NXColor)c
	{
	defaultColor = c;
	[window disableDisplay];
	[window setBackgroundColor:c];
	[scroll setColor:c];
	[window reenableDisplay];
	}


//-----------------------------------------------------------------------------
// - draggingEntered:
//-----------------------------------------------------------------------------
- (NXDragOperation) draggingEntered:(id<NXDraggingInfo>)sender
	{
	return ([sender draggingSourceOperationMask] & NX_DragOperationGeneric ?
				NX_DragOperationGeneric : NX_DragOperationNone);
	}


//-----------------------------------------------------------------------------
// - performDragOperation:
//-----------------------------------------------------------------------------
- (BOOL) performDragOperation:(id<NXDraggingInfo>)sender
	{
	[self setDaColors: NXReadColorFromPasteboard([sender draggingPasteboard])];
	[window display];
	return YES;
	}


//-----------------------------------------------------------------------------
// - initDir:
//-----------------------------------------------------------------------------
- initDir:(char const*)dirname
	{
	char buff[ FILENAME_MAX + 1 ];
	[super init];
	[[NXBundle bundleForClass:[self class]]
		getPath:buff forResource:[[self class] name] ofType:"nib"];
	[NXApp loadNibFile:buff owner:self withNames:NO fromZone:[self zone]];
	[self setDaColors: defaultColor];
	[window registerForDraggedTypes:&NXColorPboardType count:1];
	path = 0;
	showHidden = [hiddenFilesSwitch state];
	highlightDirs = [highlightSwitch state];
	[autoSortSwitch setState:[scroll autoSortRows]];
	[scroll setDoubleAction: @selector(didDoubleClick:)];
	[self load:dirname];
	[self cascade];
	[window makeKeyAndOrderFront:self];
	[openDirs addObject:self];
	return self;
	}


//-----------------------------------------------------------------------------
// - init
//-----------------------------------------------------------------------------
- init
	{
	return [self initDir: NXHomeDirectory()];
	}


//-----------------------------------------------------------------------------
// - path
//-----------------------------------------------------------------------------
- (char const*) path
	{
	return path;
	}


//-----------------------------------------------------------------------------
// + findDir:
//-----------------------------------------------------------------------------
+ (DirWindow*) findDir:(char const*) normalized_path
	{
	DirWindow* p = 0;
	char const* s;
	unsigned int i;
	unsigned int const lim = [openDirs count];
	if (normalized_path != 0)
		{
		for (i = 0;	 i < lim;  i++)
			{
			p = (DirWindow*) [openDirs objectAt:i];
			s = [p path];
			if (s != 0 && strcmp( s, normalized_path ) == 0)
				return p;
			}
		}
	return 0;
	}


//-----------------------------------------------------------------------------
// + launchDir:
//-----------------------------------------------------------------------------
+ launchDir: (char const*) dirname
	{
	DirWindow* p = 0;
	char buff[ FILENAME_MAX + 1 ];
	if (dirname == 0) dirname = NXHomeDirectory();
	if (dirname == 0) dirname = "/";
	strncpy( buff, dirname, sizeof(buff) );
	buff[ sizeof(buff) - 1 ] = '\0';
	normalize_path( buff, sizeof(buff) );
	if ((p = [self findDir:buff]) != 0)
		[p makeKeyAndOrderFront:self];
	else
		p = [[self alloc] initDir:buff];
	return p;
	}


@end
